#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2008 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Text;
using System.Xml;
using HeuristicLab.Core;

namespace HeuristicLab.Operators {
  /// <summary>
  /// Class to inject local variables into the scope.
  /// </summary>
  public class VariableInjector : OperatorBase, IVariablesContainer {
    private Dictionary<string, IVariable> myVariables;
    /// <summary>
    /// Gets a collection of all variables of the current operator.
    /// </summary>
    public virtual ICollection<IVariable> Variables {
      get { return myVariables.Values; }
    }

    /// <inheritdoc select="summary"/>
    public override string Description {
      get { return @"Injects new variable into a scope."; }
    }

    /// <summary>
    /// Overrides the Parameters property so that a parameter is returned for each injected variable.
    /// </summary>
    public override ICollection<IParameter> Parameters {
      get {
        List<IParameter> parameters = new List<IParameter>();
        foreach (IVariable variable in Variables)
          parameters.Add(new Parameter(variable.Name, "Injected variable.", variable.Value.GetType(), ParameterType.Out));
        return parameters;
      }
    }

    /// <summary>
    /// Initializes a new instance of <see cref="VariableInjector"/>.
    /// </summary>
    public VariableInjector()
      : base() {
      myVariables = new Dictionary<string, IVariable>();
    }

    /// <summary>
    /// Clones the current instance (deep clone).
    /// </summary>
    /// <remarks>Deep clone performed with <see cref="Auxiliary.Clone"/> of helper class
    /// <see cref="Auxiliary"/>.</remarks>
    /// <param name="clonedObjects">Dictionary of all already cloned objects. (Needed to avoid cycles.)</param>
    /// <returns>The cloned object as <see cref="VariableInjector"/>.</returns>
    public override object Clone(IDictionary<Guid, object> clonedObjects) {
      VariableInjector clone = (VariableInjector)base.Clone(clonedObjects);
      foreach (IVariable variable in Variables)
        clone.AddVariable((IVariable)Auxiliary.Clone(variable, clonedObjects));
      return clone;
    }

    /// <summary>
    /// Adds the specified variables to the given <paramref name="scope"/> (and removes them first,
    /// if they already exist in the current scope).
    /// </summary>
    /// <param name="scope">The scope where to inject the variables.</param>
    /// <returns><c>null</c>.</returns>
    public override IOperation Apply(IEnvironment env, IScope scope) {
      foreach (IVariable variable in Variables) {
        if (scope.GetVariable(variable.Name) != null)
          scope.RemoveVariable(variable.Name);
        scope.AddVariable((IVariable)variable.Clone());
      }
      return null;
    }

    /// <summary>
    /// Creates a new instance of <see cref="VariableInjectorView"/> to display the current instance.
    /// </summary>
    /// <returns>The created view as <see cref="VariableInjectorView"/>.</returns>
    public override IView CreateView() {
      return new VariableInjectorView(this);
    }

    /// <summary>
    /// Gets a variable with the given <paramref name="name"/>.
    /// </summary>
    /// <param name="name">The name of the variable.</param>
    /// <returns>The variable with the specified name.</returns>
    public virtual IVariable GetVariable(string name) {
      IVariable variable;
      if (myVariables.TryGetValue(name, out variable))
        return variable;
      else
        return null;
    }
    /// <summary>
    /// Adds the specified <paramref name="variable"/> to the current instance.
    /// </summary>
    /// <param name="variable">The variable to add.</param>
    public virtual void AddVariable(IVariable variable) {
      myVariables.Add(variable.Name, variable);
      variable.NameChanging += new EventHandler<NameChangingEventArgs>(Variable_NameChanging);
      variable.NameChanged += new EventHandler(Variable_NameChanged);
      OnVariableAdded(variable);
    }
    /// <summary>
    /// Deletes the variable with the specified <paramref name="name"/>.
    /// </summary>
    /// <param name="name">The name of the variable to delete.</param>
    public virtual void RemoveVariable(string name) {
      IVariable variable;
      if (myVariables.TryGetValue(name, out variable)) {
        variable.NameChanging -= new EventHandler<NameChangingEventArgs>(Variable_NameChanging);
        variable.NameChanged -= new EventHandler(Variable_NameChanged);
        myVariables.Remove(name);
        OnVariableRemoved(variable);
      }
    }
    private void Variable_NameChanging(object sender, NameChangingEventArgs e) {
      e.Cancel = myVariables.ContainsKey(e.Name);
    }
    private void Variable_NameChanged(object sender, EventArgs e) {
      IVariable variable = (IVariable)sender;
      string oldName = null;
      foreach (KeyValuePair<string, IVariable> element in myVariables) {
        if (element.Value == variable)
          oldName = element.Key;
      }
      myVariables.Remove(oldName);
      myVariables.Add(variable.Name, variable);
    }

    /// <summary>
    /// Occurs when a variable has been added.
    /// </summary>
    public event EventHandler<VariableEventArgs> VariableAdded;
    /// <summary>
    /// Fires a new <c>VariableAdded</c> event.
    /// </summary>
    /// <param name="variable">The variable that has been added.</param>
    protected virtual void OnVariableAdded(IVariable variable) {
      if (VariableAdded != null)
        VariableAdded(this, new VariableEventArgs(variable));
    }
    /// <summary>
    /// Occurs when a variable has been deleted.
    /// </summary>
    public event EventHandler<VariableEventArgs> VariableRemoved;
    /// <summary>
    /// Fires a new <c>VariableRemoved</c> event.
    /// </summary>
    /// <param name="variable">The variable that has been removed</param>
    protected virtual void OnVariableRemoved(IVariable variable) {
      if (VariableRemoved != null)
        VariableRemoved(this, new VariableEventArgs(variable));
    }

    #region Persistence Methods
    /// <summary>
    /// Saves the current instance as <see cref="XmlNode"/> in the specified <paramref name="document"/>.
    /// </summary>
    /// <remarks>
    /// Calls <see cref="OperatorBase.GetXmlNode"/> of base class <see cref="OperatorBase"/>.
    /// <br/>A quick overview how the single elements of the current instance are saved:
    /// <list type="bullet">
    /// <item>
    /// <term>Variables: </term>
    /// <description>Saved as child node with tag name <c>Variables</c>. All variables are themselves
    /// saved as child nodes.</description>
    /// </item> 
    /// </list>
    /// </remarks>
    /// <param name="name">The (tag)name of the <see cref="XmlNode"/>.</param>
    /// <param name="document">The <see cref="XmlDocument"/> where to save the data.</param>
    /// <param name="persistedObjects">The dictionary of all already persisted objects. (Needed to avoid cycles.)</param>
    /// <returns>The saved <see cref="XmlNode"/>.</returns>
    public override XmlNode GetXmlNode(string name, XmlDocument document, IDictionary<Guid, IStorable> persistedObjects) {
      XmlNode node = base.GetXmlNode(name, document, persistedObjects);
      XmlNode variablesNode = document.CreateNode(XmlNodeType.Element, "Variables", null);
      foreach (IVariable variable in myVariables.Values)
        variablesNode.AppendChild(PersistenceManager.Persist(variable, document, persistedObjects));
      node.AppendChild(variablesNode);
      return node;
    }
    /// <summary>
    /// Loads the persisted variable injector from the specified <paramref name="node"/>.
    /// </summary>
    /// <remarks>Calls <see cref="OperatorBase.Populate"/> of base class 
    /// <see cref="OperatorBase"/>.
    /// For informations how the different elements must be saved please see <see cref="GetXmlNode"/>.</remarks>
    /// <param name="node">The <see cref="XmlNode"/> where the operation is saved.</param>
    /// <param name="restoredObjects">A dictionary of all already restored objects. (Needed to avoid cycles.)</param>
    public override void Populate(XmlNode node, IDictionary<Guid, IStorable> restoredObjects) {
      base.Populate(node, restoredObjects);
      XmlNode variablesNode = node.SelectSingleNode("Variables");
      foreach (XmlNode variableNode in variablesNode.ChildNodes)
        AddVariable((IVariable)PersistenceManager.Restore(variableNode, restoredObjects));
    }
    #endregion
  }
}
